#!/usr/bin/env python3

from ansible.module_utils.basic import AnsibleModule

ANSIBLE_METADATA = {
    'metadata_version': '0.0.1',
    'status': ['preview'],
    'supported_by': 'dev@mirantis.com'
}

DOCUMENTATION = r'''
---
author:
 - Mirantis (dev@mirantis.com)
module: eparted
short_description: Configure block device partitions
version_added: "0.0.1"
description:
  - 'This module allows configuring block device partition using the C(parted)
    command line tool. For a full description of the fields and the options
    check the GNU parted manual.'
requirements:
  - This module requires parted version 1.8.3 and above in ramfs
  - ['parted', 'lvm2', 'multipath-tools', 'lsscsi']
options:
  target_storage:
    required: False
    description: The block device (disk) where to operate.
    type: list
    required: False
  post_deploy_script:
    type: str
  target_system:
    type: dict
'''

def run_command(cmd, fail_msg='', ok_exit_codes=(0, )):
    global module

    exit_code, stdout, stderr = module.run_command(cmd)
    if exit_code not in ok_exit_codes:
        module.fail_json(msg=f"{fail_msg} "
                             f"exit_code: {exit_code} "
                             f"stdout: {stdout} "
                             f"stderr: {stderr}")

    return exit_code, stdout, stderr


def wait_for_blockdev_rereadpt(device):
    """
    re-read disk partitions
    :param device: /dev/vda
    :param attempts:
    :return:
    """

    run_command(['blockdev', '--rereadpt', device])
    udevadm_trigger_blocks()


def udevadm_trigger_blocks():
    run_command(['udevadm', 'trigger', '--subsystem-match=block'])
    wait_for_udev_settle()


def wait_for_udev_settle():
    """Wait for emptiness of udev queue within attempts*0.1 seconds"""
    run_command(['udevadm', 'settle', '--timeout=60'])


def run_module():
    global module

    changed = False
    module = AnsibleModule(
        argument_spec=dict(
            blockdev_rereadpt=dict(type='dict'),
        ),
        supports_check_mode=False,
    )
    module.run_command_environ_update = {
        'LANG': 'C',
        'LC_ALL': 'C',
        'LC_MESSAGES': 'C',
        'LC_CTYPE': 'C'
    }

    blockdev_rereadpt = module.params['blockdev_rereadpt']

    if blockdev_rereadpt:
        wait_for_blockdev_rereadpt(blockdev_rereadpt['device'])
        changed = True
    else:
        module.fail_json(msg=f'Wrong module invocation')

    module.exit_json(
        changed=changed
    )


if __name__ == '__main__':
    run_module()
