#!/usr/bin/env python3
# Copyright 2020 Mirantis, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ansible.module_utils.basic import AnsibleModule

import json
import re
import os
import stat
import traceback


ANSIBLE_METADATA = {
    'metadata_version': '0.0.1',
    'status': ['preview'],
    'supported_by': 'dev@mirantis.com'
}

DOCUMENTATION = r'''
---
author:
 - Mirantis (dev@mirantis.com)
module: block_device
short_description: Get block device(s)
version_added: "0.0.1"
description:
  - 'lsblk helper'
requirements:
  - ['util-linux']
'''


def run_command(cmd, fail_msg='', ok_exit_codes=(0, )):
    global module

    exit_code, stdout, stderr = module.run_command(cmd)
    if exit_code not in ok_exit_codes:
        module.fail_json(
            msg="{} cmd: {} exit_code: {} stdout: {} stderr: {}"
                .format(fail_msg, cmd, exit_code, stdout, stderr))

    return exit_code, stdout, stderr


def detect_partitions(block_devices):
    """
    'lsblk' returns list of structures that describe a given device or all
    devices available. Those structures may conintans nested structures under
    'children' key. A structure describes either entire device or its
    partition or logical volume (e.g. LVM or MD).
    There is 'type' field that indicates record type, but its value
    is incorrect if a record describes partition on a loopback device.
    In this case its value is 'loop' even if there is partition metadata.
    Another problem with 'lsblk' is that there is no partition number in
    its output.
    This function tries to detect if device record describes partition,
    and tries to get correct partition number.
    For each record it adds 'partnum' field that is either a partition number
    (an integer) or None if it's not a partition.

    :param block_devices: list of dicts each of which is a structure returned
                          by 'lsblk' that describes a device. It may have
                          nested structures under 'children' key.
    """
    for device in block_devices:
        if 'children' in device:
            # Detect partitions in nested structures
            detect_partitions(device['children'])

        kname = device["kname"]
        pkname = device["pkname"]
        if kname and pkname:
            partnum = get_partition_number(pkname, kname)
        else:
            partnum = None

        if partnum is None:
            device['partnum'] = None
            device['is_partition'] = False
        else:
            device['partnum'] = partnum
            device['is_partition'] = True


def get_sysfs_path(device_path, parent_device_path=None):
    """
    Returns sysfs path of a given device path. If optional parent device path
    is given then device_path must be a partition of that parent device.
    Before returning sysfs path it is checked it it's exist.

    :param device_path:
    :param parent_device_path:
    :return: sysfs path or None (if path does not exist)
    """
    device_name = os.path.split(device_path)[-1]
    sysfs_path = os.path.join("/sys/block", device_name)
    if os.path.isdir(sysfs_path):
        return sysfs_path

    if parent_device_path:
        parent_device_name = os.path.split(parent_device_path)[-1]
        sysfs_path = os.path.join("/sys/block",
                                  parent_device_name, device_name)
        if os.path.isdir(sysfs_path):
            return sysfs_path


def detect_devices_type(block_devices):
    """
    This function iterates over passed block devices tree structure
    and adds boolean flags that identifies device as MD or device mapper.

    This function helps to identify a device type with a single flag
    because 'TYPE' field returned by 'lsblk' may have different values:
    * For MD device:
      - string value from 'md/level' if exists. In this case it indicates
        RAID level ('raid0', 'raid1', 'raid5' and so on)
      - 'md' string otherwise.
    * For device mapper device:
      - UUID prefix from 'dm/uuid', e.g.
        'lvm', 'crypt', 'dmraid', 'mpath', 'part'
      - 'dm' string otherwise

    :param block_devices: tree of block devices
    :return: None
    """
    for block_device in block_devices:
        if 'children' in block_device:
            detect_devices_type(block_device['children'])

        name = os.path.split(block_device['kname'])[-1]
        sysfs_path = get_sysfs_path(device_path=block_device['kname'],
                                    parent_device_path=block_device['pkname'])

        if sysfs_path:
            block_device['is_md_device'] = \
                name.startswith('md') and \
                os.path.isdir(os.path.join(sysfs_path, 'md'))

            block_device['is_device_mapper'] = \
                name.startswith('dm-') and \
                os.path.isdir(os.path.join(sysfs_path, 'dm'))
        else:
            block_device['is_md_device'] = False
            block_device['is_device_mapper'] = False


def probe_filesystem_uuid(device_path):
    """
    Returns low-level filesystem uuid, if it is exist.

    :param device_path: Absolute path to block device.
    :return: Filesystem's UUID or None
    """
    if not device_path:
        raise Exception("Bad device name / path: '{}'".format(device_path))

    cmd = ['blkid', '--probe', '--match-tag=UUID', '--output=value',
           device_path]
    exit_code, stdout, stderr = run_command(cmd)
    if exit_code != 0:
        raise Exception("Command '{}' failed".format(' '.join(cmd)))
    for line in stdout.splitlines():
        line = line.strip()
        if line:
            return line
    return None


def confirm_filesystem_uuid(block_devices):
    """
    Verifies that filesystem UUID returned by 'lsblk' is the same as
    that filesystem has.
    This is useful when you want to get 'lsblk' output immediately after
    creating new filesystem on a device. In that case 'lsblk' may return
    UUID of a filesystem that exists *before* altering filesystem, because
    'lsblk' doesn't get data directly from device.
    To workaround this 'blkid --probe' might be used. This will return
    UUID by probing filesystem, so if both UUIDs match we might be sure
    that 'lsblk' data is correct.
    One of the usecases for this is to verify filesystem UUID when
    looking up for a specific partition. In that case the returned partition
    *should* include valid filesystem UUID.
    :param block_devices:
    :return:
    """
    global module

    for device in block_devices:
        if 'children' in device:
            confirm_filesystem_uuid(device['children'])

        if not device.get('is_partition', False):
            continue

        device_uuid = probe_filesystem_uuid(device['kname'])
        if device['uuid'] != device_uuid:
            module.fail_json(msg="Device UUID mismatch for device '{}'. "
                                 "lsblk UUID: '{}', blkid UUID: '{}'"
                             .format(device['kname'],
                                     device['uuid'],
                                     device_uuid))


def get_partition_number(parent_device_path, device_path):
    """
    Gets partition number for a device (identified by its absolute path)
    that is a child of another device (also identified by its absolute path).

    :param parent_device_path: Absolute path to the device that holds
                               the partition.
    :param device_path: Absolute path to the device that is a partition.
    :return: partition number or None.
    """

    if not (parent_device_path and device_path):
        raise Exception("Both 'device_path' and 'parent_device_path' should "
                        "be a path to device, got 'device': '{}', "
                        "'parent_device': '{}'".format(device_path,
                                                       parent_device_path))

    sysfs_path = get_sysfs_path(device_path=device_path,
                                parent_device_path=parent_device_path)
    partition_file = os.path.join(sysfs_path, 'partition')

    if os.path.exists(partition_file):
        with open(partition_file) as f:
            return int(f.read())


def resolve_device_path(device_path):
    """
    :param device_path: Device path, absolute or relative to '/dev'
                        or '/dev/mapper'.
    :return: Full device path if block device exists, None otherwise.
    """
    paths = []
    if os.path.isabs(device_path):
        paths.append(os.path.realpath(device_path))
    else:
        for prefix in ['/dev', '/dev/mapper']:
            paths.append(os.path.realpath(os.path.join(prefix, device_path)))

    for path in paths:
        if os.path.exists(path) and stat.S_ISBLK(os.stat(path).st_mode):
            return path

    return None


def resolve_device_paths(block_devices):
    """
    Resolves device-related paths stored as 'kname' and 'pkname' to their
    real absolute paths for every device in the structure.
    Result should be the same, actually, as 'k' suffix is for 'kernel' and all
    those paths should already be absolute paths under '/dev'.

    :param block_devices: tree-like structure of block devices
    :return: None, updated values added to the original structure.
    """
    for block_device in block_devices:
        for key in ['kname', 'pkname']:
            if key in block_device:
                if block_device[key]:
                    block_device[key] = resolve_device_path(block_device[key])

        if 'children' in block_device:
            resolve_device_paths(block_device['children'])


def lvm_report():
    cmd = [
        "lvm", "fullreport",
        "--reportformat", "json",
        "--units", "b",
        "--nosuffix",
        "--configreport", "lv",
        "--options", "lv_name,lv_size,vg_name,lv_uuid,lv_path",
        "--configreport", "vg",
        "--options", "vg_name,vg_uuid,vg_size,vg_free",
        "--configreport", "pv",
        "--options", "pv_name,vg_name,pv_size,dev_size,pv_uuid"
    ]
    _, stdout, _ = run_command(
        cmd, fail_msg="'{}' failed.".format(cmd))

    lvs = {}
    pvs = {}
    vgs = {}

    for item in json.loads(stdout)["report"]:
        for lv in item.get("lv", []):
            lv_uuid = lv["lv_uuid"]
            if lv_uuid not in lvs:
                lvs[lv_uuid] = lv

        for pv in item.get("pv", []):
            pv_uuid = pv["pv_uuid"]
            if pv_uuid not in pvs:
                pvs[pv_uuid] = pv

        for vg in item.get("vg", []):
            vg_uuid = vg["vg_uuid"]
            if vg_uuid not in vgs:
                vgs[vg_uuid] = vg

    for lv in lvs.values():
        lv["lv_kname"] = os.path.realpath(lv["lv_path"])

    for pv in pvs.values():
        pv["pv_kname"] = os.path.realpath(pv["pv_name"])

    return {
        "lv": list(lvs.values()),
        "pv": list(pvs.values()),
        "vg": list(vgs.values())
    }


def insert_lvm_data(block_devices, lvm_data):
    for block_device in block_devices:
        if "children" in block_device:
            insert_lvm_data(block_device["children"], lvm_data)

        block_device.update({
            "lv_name": "",
            "lv_uuid": "",
            "lv_path": "",
            "vg_name": "",
            "pv_name": "",
            "pv_uuid": "",
        })

        kname = block_device["kname"]
        lv = next(
            filter(lambda x: x["lv_kname"] == kname, lvm_data["lv"]), None)

        if lv:
            block_device.update({
                "lv_name": lv["lv_name"],
                "lv_uuid": lv["lv_uuid"],
                "lv_path": lv["lv_path"],
                "vg_name": lv["vg_name"],
            })
        else:
            pv = next(
                filter(lambda x: x["pv_kname"] == kname, lvm_data["pv"]), None)

            if pv:
                block_device.update({
                    "pv_name": pv["pv_name"],
                    "pv_uuid": pv["pv_uuid"],
                    "vg_name": pv["vg_name"],
                })


def get_block_devices(device_name, device_match, custom_fields='',
                      get_filesystem=False, get_partitions=False,
                      get_lvm=False):
    lsblk_fields = {'NAME', 'KNAME', 'PKNAME', 'SIZE', 'TYPE'}

    if get_filesystem:
        lsblk_fields |= {'FSTYPE', 'LABEL', 'UUID', 'MOUNTPOINT'}

    if get_partitions:
        lsblk_fields |= {'PARTTYPE', 'PARTLABEL', 'PARTUUID', 'PARTFLAGS'}

    for field in custom_fields.split(','):
        if field:
            lsblk_fields.add(field.strip().upper())

    for key in device_match:
        if key in ['device', ]:
            continue
        lsblk_fields.add(key.upper())

    cmd = ['lsblk', '--json', '--bytes', '--paths',
           '--output', ','.join(lsblk_fields)]

    if device_name:
        cmd.append(device_name)

    exit_code, stdout, stderr = run_command(cmd,
                                            fail_msg="'lsblk' failed.")
    block_devices = json.loads(stdout)['blockdevices']

    if device_match:
        if 'device' in device_match:
            name = device_match['device']
            if name.startswith('^'):
                re_name = re.compile(name)
                block_devices = list(filter(
                    lambda x: re_name.match(x['name']), block_devices
                ))
            else:
                block_devices = list(filter(
                    lambda x: x['name'].startswith(name), block_devices
                ))
        for key, value in device_match.items():
            if key in ['device', ]:
                continue
            block_devices = list(filter(
                lambda x: x[key.lower()] == value, block_devices
            ))

    resolve_device_paths(block_devices)

    detect_devices_type(block_devices)

    if get_partitions:
        detect_partitions(block_devices)

    if get_lvm:
        lvm_data = lvm_report()
        insert_lvm_data(block_devices, lvm_data)

    return block_devices


def run_module():
    global module

    changed = False
    module = AnsibleModule(
        argument_spec=dict(
            device=dict(type='str'),
            match=dict(type='dict', default={}),
            partnum=dict(type='int'),
            filesystem=dict(type='bool', default=False),
            partitions=dict(type='bool', default=False),
            lvm=dict(type='bool', default=False),
            fields=dict(type='str', default=''),
            confirm_uuid=dict(type='bool', default=False),
            state=dict(type='str'),
        ),
        supports_check_mode=False,
    )
    module.run_command_environ_update = {
        'LANG': 'C',
        'LC_ALL': 'C',
        'LC_MESSAGES': 'C',
        'LC_CTYPE': 'C'
    }

    device_name = module.params['device']
    device_match = module.params['match']
    partition_number = module.params['partnum']
    confirm_uuid = module.params['confirm_uuid']
    custom_fields = module.params['fields']
    get_filesystem = module.params['filesystem']
    get_partitions = module.params['partitions']
    get_lvm = module.params['lvm']

    if partition_number is not None:
        if not device_name:
            module.fail_json(msg="'partnum' requires 'device' parameter.")

        if device_match:
            module.fail_json(msg="'partnum' can't be used with 'match', "
                                 "only 'device' is supported.")

        get_filesystem = True
        get_partitions = True

    if confirm_uuid:
        get_filesystem = True
        get_partitions = True

    try:
        block_devices = get_block_devices(device_name, device_match,
                                          custom_fields,
                                          get_filesystem, get_partitions,
                                          get_lvm)
    except: # noqa
        block_devices = []
        module.fail_json(
            msg="Failed to get list of block devices:\n{}"
                .format(traceback.format_exc()),
        )

    if partition_number is not None:
        partitions = []
        for device in block_devices:
            for partition in device.get('children', []):
                if partition.get('partnum') == partition_number:
                    partitions.append(partition)

        confirm_filesystem_uuid(partitions)

        module.exit_json(
            changed=changed,
            blockdevices=partitions,
        )
    else:
        if confirm_uuid:
            confirm_filesystem_uuid(block_devices)

        module.exit_json(
            changed=changed,
            blockdevices=block_devices,
        )


if __name__ == '__main__':
    run_module()
